@@ -6,6 +6,8 @@ import atexit
 import concurrent.futures
 from dataclasses import dataclass, field
 from enum import Enum
+from functools import wraps
+import json
 
 import logging
 import os
@@ -13,17 +15,33 @@ import pkg_resources
 import sys
 import time
 import traceback
-from typing import List, Optional
+from typing import Any, Callable, Dict, List, Optional
 import uuid
 
-import requests
+import requests  # type: ignore
 
+
+from . import utils
 from .consent import HumbugConsent
 from .system_information import (
     SystemInformation,
     generate as generate_system_information,
 )
 
+psutil = None
+
+try:
+    import psutil  # type: ignore
+except ImportError:
+    pass
+
+GPUtil = None
+
+try:
+    import GPUtil  # type: ignore
+except ImportError:
+    pass
+
 
 DEFAULT_URL = "https://spire.bugout.dev"
 
@@ -58,6 +76,8 @@ class HumbugReporter:
         timeout_seconds: int = 10,
         mode: Modes = Modes.DEFAULT,
         url: Optional[str] = None,
+        tags: Optional[List[str]] = None,
+        blacklist_fn: Optional[Callable[[Dict[str, Any]], Dict[str, Any]]] = None,
     ):
         if url is None:
             url = DEFAULT_URL
@@ -87,6 +107,15 @@ class HumbugReporter:
         self.is_excepthook_set = False
         self.is_loggerhook_set = False
 
+        self.tags: List[str] = []
+        if tags is not None:
+            self.tags = tags
+
+        self.blacklist_fn = blacklist_fn
+
+        self.psutil_exists = psutil is not None
+        self.gputil_exists = GPUtil is not None
+
     def wait(self) -> None:
         concurrent.futures.wait(
             self.report_futures, timeout=float(self.timeout_seconds)
@@ -113,13 +142,20 @@ class HumbugReporter:
 
         return tags
 
+    def _post_body(self, report: Report) -> Dict[str, Any]:
+        return {
+            "title": report.title,
+            "content": report.content,
+            "tags": report.tags + self.tags,
+        }
+
     def publish(self, report: Report, wait: bool = False) -> None:
         if not self.consent.check():
             return
         if self.bugout_token is None:
             return
 
-        json = {"title": report.title, "content": report.content, "tags": report.tags}
+        json = self._post_body(report)
         headers = {
             "Authorization": "Bearer {}".format(self.bugout_token),
         }
@@ -140,7 +176,7 @@ class HumbugReporter:
                     timeout=self.timeout_seconds,
                 )
                 self.report_futures.append(report_future)
-        except:
+        except Exception:
             pass
 
     def custom_report(
@@ -237,7 +273,14 @@ Release: `{os_release}`
         )
         if tags is None:
             tags = []
-        tags.append("type:error")
+
+        tags.extend(["type:error", "error:{}".format(error.__class__.__name__)])
+        try:
+            tags.append(
+                "error_full:{}.{}".format(error.__module__, error.__class__.__name__),
+            )
+        except Exception:
+            pass
         tags.extend(self.system_tags())
 
         report = Report(title=title, content=error_content, tags=tags)
@@ -359,6 +402,93 @@ Release: `{os_release}`
 
         return report
 
+    def feature_report(
+        self,
+        feature_name: str,
+        parameters: Dict[str, Any],
+        tags: Optional[List[str]] = None,
+        publish: bool = True,
+        wait: bool = False,
+        apply_blacklist: bool = True,
+    ) -> Report:
+        title = "Feature used: {name}".format(name=feature_name)
+
+        if apply_blacklist and self.blacklist_fn is not None:
+            parameters = self.blacklist_fn(parameters)
+
+        parameters_content = "\n".join(
+            [
+                "- `{parameter_name}` = `{parameter_value}`".format(
+                    parameter_name=key, parameter_value=value
+                )
+                for key, value in parameters.items()
+            ]
+        )
+
+        content = """### User timestamp
+```
+{user_time}
+```
+
+### Information
+
+Feature: {name}
+
+{parameters_content}
+""".format(
+            user_time=int(time.time()),
+            name=feature_name,
+            parameters_content=parameters_content,
+        )
+
+        if tags is None:
+            tags = []
+        tags.append("type:feature")
+        tags.append("feature:{}".format(feature_name))
+        tags.extend(self.system_tags())
+        tags.extend(
+            ["parameter:{}={}".format(key, value) for key, value in parameters.items()]
+        )
+
+        report = Report(title=title, content=content, tags=tags)
+
+        if publish:
+            self.publish(report, wait=wait)
+
+        return report
+
+    def record_call(
+        self,
+        callable: Callable,
+    ) -> Callable:
+        @wraps(callable)
+        def wrapped_callable(*args, **kwargs):
+            parameters = {**kwargs}
+            for i, arg in enumerate(args):
+                parameters["arg.{}".format(i)] = arg
+
+            self.feature_report(callable.__name__, parameters)
+
+            return callable(*args, **kwargs)
+
+        return wrapped_callable
+
+    def record_errors(
+        self,
+        callable: Callable,
+    ) -> Callable:
+        @wraps(callable)
+        def wrapped_callable(*args, **kwargs):
+            result = None
+            try:
+                result = callable(*args, **kwargs)
+            except Exception as err:
+                self.error_report(err, tags=["site:{}".format(callable.__name__)])
+                raise err
+            return result
+
+        return wrapped_callable
+
     def setup_loggerhook(
         self,
         level: int,
@@ -414,6 +544,65 @@ Release: `{os_release}`
         ipython_shell.showtraceback = showtraceback
         self.setup_excepthook(publish=True, tags=tags)
 
+    def metrics_report(
+        self,
+        cpu: bool = True,
+        gpu: bool = True,
+        memory: bool = True,
+        disk: bool = True,
+        network: bool = True,
+        open_files_flag: bool = True,
+        num_threads_flag: bool = True,
+        processes_flag: bool = True,
+        tags: Optional[List[str]] = None,
+        publish: bool = False,
+        wait: bool = False,
+    ) -> Report:
+        title = "Metrics report"
+
+        metrics: Dict[str, Any] = {}
+
+        if self.gputil_exists:
+            metrics["gpu"] = utils.get_gpu_metrics()
+
+        if self.psutil_exists:
+            if cpu:
+                metrics["cpu"] = utils.get_cpu_metrics()
+
+            if memory:
+                metrics["memory"] = utils.get_memory_metrics()
+
+            if disk:
+                metrics["disk"] = utils.get_disk_metrics()
+
+            if network:
+                metrics["network"] = utils.get_network_metrics()
+
+            if open_files_flag:
+                metrics["open_files"] = utils.get_open_files_metrics()
+
+            if num_threads_flag:
+                metrics["num_threads"] = utils.get_thread_metrics()
+
+            if processes_flag:
+                metrics["processes"] = utils.get_processes_metrics()
+
+        tags = tags if tags is not None else []
+
+        tags.append("type:metrics")
+        tags.extend(self.system_tags())
+
+        report = Report(
+            title=title,
+            tags=tags,
+            content=f"```\n{json.dumps(metrics, indent=4, sort_keys=True)}\n```",
+        )
+
+        if publish:
+            self.publish(report, wait=wait)
+
+        return report
+
 
 class Reporter(HumbugReporter):
     """
@@ -484,5 +673,5 @@ class Reporter(HumbugReporter):
                     timeout=self.timeout_seconds,
                 )
                 self.report_futures.append(report_future)
-        except:
+        except Exception:
             pass
